#ifndef __FWEELIN_CORE_H
#define __FWEELIN_CORE_H

/*
   ********
   Art is what's going on right under our noses.
   An artist is just one who listens to it.
   ********
*/

#include "fweelin_midiio.h"
#include "fweelin_videoio.h"
#include "fweelin_keyio.h"
#include "fweelin_audioio.h"

#include "fweelin_core_dsp.h"
#include "fweelin_block.h"
#include "fweelin_event.h"
#include "fweelin_config.h"

class KeyIO;
class EventManager;
class BlockManager;
class RootProcessor;
class RecordProcessor;
class TriggerMap;
class AudioBlock;
class AudioBlockIterator;
class Loop;
class Pulse;
class BED_MarkerPoints;
class VorbisStreamer;
class PreallocatedType;
class AudioBuffers;
class InputSettings;

#if USE_FLUIDSYNTH
class FluidSynthProcessor;
#endif

// ****************** CORE CLASSES

// Loop 
// -Wraps around a list of audio blocks
// -Stores basic parameters for a loop, such as volumes
class Loop {
public:
  const static float MIN_VOL = 0.01;

  Loop (AudioBlock *blocks, Pulse *pulse, float quant, float vol,
	long nbeats) : 
    blocks(blocks), pulse(pulse), quant(quant), 
    vol(vol), dvol(1.0), saved(0), nbeats(nbeats) {};

  // RT update volume
  inline void UpdateVolume() {
    if (dvol != 1.0) {
      // Apply delta
      if (dvol > 1.0 && vol < MIN_VOL)
	vol = MIN_VOL;
      vol *=dvol;
    }
  };

  // Perhaps store the length of the loop...
  // Because right now GetTotalLength is a calculation
  // in AudioBlock

  AudioBlock *blocks; // Chain of blocks that form this loop
  
  // Time pulse to which this loop is quantized
  // And quantization factor.
  Pulse *pulse;
  float quant;

  float vol,   // Volume multiplier
    dvol;      // Rate of volume change

  char saved;  // Nonzero if this loop has been saved already
    
  long nbeats; // Number of beats in this loop
};

// This class manages sets of loops, allowing them to be triggered..
// like a keymap
class TriggerMap {
public:
  TriggerMap (int mapsize) : mapsize(mapsize), lastupdate(0) {
    // Map is an array of pointers to different loops
    map = new (Loop *)[mapsize];
    memset(map,0,mapsize * sizeof(Loop *));
  };

  ~TriggerMap () {
    delete[] map;
  }

  void SetMap (int index, Loop *smp) {
    if (index < 0 || index >= mapsize) {
      printf("SetMap: Invalid loop index!\n");
    }
    else {
      map[index] = smp;
      TouchMap();
    }
  }

  Loop *GetMap (int index) {
    if (index < 0 || index >= mapsize) {
      printf("GetMap: Invalid loop index!\n");
      return 0;
    }
    else
      return map[index];
  }

  int GetMapSize() { return mapsize; }
  double GetLastUpdate() { return lastupdate; }
  void TouchMap() { lastupdate = mygettime(); };

protected:

  int mapsize;
  Loop **map;

  // Time of last update of the map- used for rebuilding tables & such
  double lastupdate;
};

// Status types for loops
enum LoopStatus {
  T_LS_Off,

  T_LS_Recording,
  T_LS_Overdubbing,
  T_LS_Playing
};

class LoopList {
 public:
  LoopList(Loop *l = 0, int l_idx = 0) : l(l), l_idx(l_idx), next(0) {};

  static void Delete(LoopList *first) {
    LoopList *cur = first;
    while (cur != 0) {
      LoopList *tmp = cur->next;
      delete cur;
      cur = tmp;
    }
  }

  static void Add(LoopList **first, LoopList *nw) {
    if (*first == 0) 
      *first = nw;
    else {
      nw->next = *first;
      *first = nw;
    }
  };

  static void Remove(LoopList **first, LoopList *prev, LoopList **cur) {
    LoopList *tmp = (*cur)->next;
    if (prev != 0)
      prev->next = tmp;
    else 
      *first = tmp;
    delete (*cur);
    *cur = tmp;
  };

  Loop *l;   // Loop 
  int l_idx; // Index of loop in map

  LoopList *next;
};

// LoopManager contains all loops, and wraps up recording, playing, and
// other RT & non-RT processing on loops
class LoopManager : public EventListener, public AutoVorbisControl {
public:
  LoopManager (Fweelin *app);
  virtual ~LoopManager();

  virtual void ReceiveEvent(Event *ev, EventProducer *from);

  // Get length returns the length of any loop on the specified index
  nframes_t GetLength(int index);

  // Get length returns the length of any loop on the specified index
  // Rounded to its currently quantized length
  // Or 0 if the loop has no pulse
  nframes_t GetRoundedLength(int index);

  // Returns from 0.0-1.0 how far through a block chain with specified index
  // we are
  float GetPos(int index);

  // Get current # of samples into block chain with given index
  nframes_t GetCurCnt(int index);

  // Returns nonzero if there is a processor at the specified mapindex
  int IsActive(int index) { return (plist[index] != 0); }

  // Sets trigger volume on specified index
  // If index is not playing, activates the index
  void SetTriggerVol(int index, float vol);

  // Gets triggered volume on specified index
  // If index is not playing, returns 0
  float GetTriggerVol(int index);

  // Returns the status at the specified index
  LoopStatus GetStatus(int index) { return status[index]; };

  Processor *GetProcessor(int index) { return plist[index]; };

  // Returns a loop with the specified index, if one exists
  Loop *GetSlot(int index);

  void AdjustNewLoopVolume(float adjust) {
    newloopvol += adjust;
    if (newloopvol < 0.0)
      newloopvol = 0.0;
  }
  float GetNewLoopVolume() { return newloopvol; }

  void AdjustOutputVolume(float adjust);
  void SetOutputVolume(float set);
  float GetOutputVolume();

  void AdjustInputVolume(float adjust);
  void SetInputVolume(float set);
  float GetInputVolume();

  void SetLoopVolume(int index, float val);
  float GetLoopVolume(int index);
  void AdjustLoopVolume(int index, float adjust);
  void SetLoopdVolume(int index, float val);
  float GetLoopdVolume(int index);

  void SelectPulse(int pulseindex);

  // Create a time pulse around the specified index
  // The length of the loop on the specified index becomes
  // a time constant around which other loops center themselves
  // subdivide the length of the loop by subdivide to get the core pulse
  void CreatePulse(int index, int pulseindex, int sub);

  // Taps a pulse- starting at the downbeat- if newlen is nonzero, the pulse's
  // length is adjusted to reflect the length between taps- and a new pulse
  // is created if none exists
  void TapPulse(int pulseindex, char newlen);

  void SwitchMetronome(int pulseindex, char active);

  Pulse *GetCurPulse() {
    if (curpulseindex >= 0) 
      return pulses[curpulseindex];
    else
      return 0;
  };

  int GetCurPulseIndex() { return curpulseindex; };
  Pulse *GetPulseByIndex(int pulseindex) { return pulses[pulseindex]; };

  // Deletes the specified pulse, stopping striping
  void DeletePulse(int pulseindex);

  // Gets the pulse to which the loop on index is attached
  Pulse *GetPulse(int index);

  void SetSubdivide(int sub) {
    //printf("Set subdivide: %d\n", sub);
    subdivide = sub;
  }
  int GetSubdivide() { return subdivide; };

  // Move the loop at specified index to another index
  // only works if target index is empty
  // returns 1 if success
  int MoveLoop (int src, int tgt);

  // Delete the loop at the specified index..
  void DeleteLoop (int index);

  // Trigger the loop at index within the map
  // The exact behavior varies depending on what is already happening with
  // this loop and the settings passed- see .fweelin.rc
  void Activate (int index, float vol = 1.0, nframes_t ofs = 0, 
		 char overdub = 0, float od_feedback = 0.5);

  void Deactivate (int index);

  // Saves the loop with given index
  void SaveLoop (int index);

  // Opens the right output file and gets audio blocks
  // to save given the loop id
  void SetupSave(int index, FILE **out, AudioBlock **b, 
		 AudioBlockIterator **i, nframes_t *len);

  // We receive calls periodically for autosaving of loops-
  // here, we return blocks to save from loops which need saving
  virtual void GetBlock(FILE **out, AudioBlock **b, 
			AudioBlockIterator **i, nframes_t *len);

  // Check if the needs_saving map is up to date, rebuild if needed.
  void CheckSaveMap();

  LoopList *needs_saving;    // List of loops that need saving
  double needs_saving_stamp; // Timestamp from trigger map from which 
                             // needs_saving was built

  int *lastrecidx;   // List of last indexes recorded to
  int numloops,      // Total number of loops in map
    numrecordingloops; // Number of loops currently recording in map
  
 protected:

  // Setup time marker striping on audio memory when a new
  // pulse is selected
  void StripePulseOn(Pulse *pulse);
  
  // Deactivate time marker striping on audio memory for the
  // given pulse
  void StripePulseOff(Pulse *pulse);

  // Turn on/off auto loop saving
  void SetAutoLoopSaving(char save);

  // Core Fweelin app on which this player is working on 
  Fweelin *app;

  // We also need to keep track of record/playprocessors attached to each
  // index
  Processor **plist;  // Pointer to a processor for each index
  // *** Replace this with well designed event queue system
  int *waitactivate;  // For each index, are we waiting to activate?
  float *waitactivate_vol;
  char *waitactivate_od;
  float *waitactivate_od_fb;
  LoopStatus *status; // For each index, what's the status?

  // Block manager that auto saves loops
  VorbisWriteManager *vw;

  // Initial volume of new loops
  float newloopvol;

  // Subdivision for creating new pulses
  int subdivide;
  // Index of last deactivated trigger index.. used when creating new pulses
  int lastindex;
  // Current pulse (index) for new loops.. will be quantized to this pulse
  int curpulseindex;
  // List of possible pulses
  Pulse *pulses[MAX_PULSES];
};

// This is Fwheelin
class Fweelin : public EventProducer {
 public:

  Fweelin() : 

#if USE_FLUIDSYNTH
    fluidp(0), 
#endif

    mmg(0), bmg(0), emg(0), rp(0), tmap(0), 
    loopmgr(0), abufs(0), iset(0), audio(0), midi(0), keys(0), vid(0),
    scope(0), scope_len(0), audiomem(0), amrec(0),  
    running(0) {};
  ~Fweelin() {};

  char IsRunning() { return running; };

  // Setup flomonkey
  int setup();
  // Start flomonkey
  int go();

  void ToggleDiskOutput();
  void FlushWriteName() { strcpy(writename,""); };

  nframes_t getBUFSZ() { return fragmentsize; };
  sample_t *getSCOPE() { return scope; };
  nframes_t getSCOPELEN() { return scope_len; };

#if USE_FLUIDSYNTH
  FluidSynthProcessor *getFLUIDP() { return fluidp; };
  FluidSynthProcessor *fluidp;
#endif

  AudioBlock *getAUDIOMEM() { return audiomem; };
  AudioBlockIterator *getAUDIOMEMI();
  RecordProcessor *getAMREC() { return amrec; };
  AudioBlock *getAMPEAKS();
  AudioBlock *getAMAVGS();
  AudioBlockIterator *getAMPEAKSI();
  AudioBlockIterator *getAMAVGSI();
  BED_MarkerPoints *getAMPEAKSPULSE();

  MemoryManager *getMMG() { return mmg; };
  BlockManager *getBMG() { return bmg; };
  EventManager *getEMG() { return emg; };
  RootProcessor *getRP() { return rp; };
  VorbisStreamer *getSTREAMER() { return vs; };
  TriggerMap *getTMAP() { return tmap; };
  LoopManager *getLOOPMGR() { return loopmgr; };
  
  AudioIO *getAUDIO() { return audio; };
  MidiIO *getMIDI() { return midi; };
  VideoIO *getVIDEO() { return vid; };
  KeyIO *getKEYS() { return keys; };

  FloConfig *getCFG() { return cfg; };

  char *getOUTPUTNAME() { return writename; };

  PreallocatedType *getPRE_AUDIOBLOCK() { return pre_audioblock; };
  PreallocatedType *getPRE_TIMEMARKER() { return pre_timemarker; };

  AudioBuffers *getABUFS() { return abufs; };
  InputSettings *getISET() { return iset; };

  int GetNextOutputFolder();
  int GetOutputFolder();

 private:
  MemoryManager *mmg;
  BlockManager *bmg;
  EventManager *emg;
  RootProcessor *rp;
  TriggerMap *tmap;
  LoopManager *loopmgr;

  // ****************** PREALLOCATED TYPE MANAGERS
  PreallocatedType *pre_audioblock,
    *pre_timemarker;

  // ****************** VORBIS DISK STREAMER
  VorbisStreamer *vs;
  int writenum; // Number of ogg file currently being written
  char writename[FWEELIN_OUTNAME_LEN], // Name of ogg file
    timingname[FWEELIN_OUTNAME_LEN]; // Name of timing stripe file

  // ****************** SYSTEM LEVEL AUDIO  
  // Audio buffers
  AudioBuffers *abufs;
  // Input settings
  InputSettings *iset;
  // Audio interface
  AudioIO *audio;

  // ****************** MIDI
  MidiIO *midi;
  
  // ****************** KEYBOARD
  KeyIO *keys;
  
  // ****************** VIDEO
  VideoIO *vid;

  // Audio buffer size
  nframes_t fragmentsize;

  // Buffers for visual sample scope
  sample_t *scope;
  nframes_t scope_len;

  // Audio memory & the processor which records to it
  AudioBlock *audiomem;
  RecordProcessor *amrec;

  // Control settings 
  FloConfig *cfg;
  
  char running; // Nonzero if flomonkey is actually running
};

#endif
