#if USE_FLUIDSYNTH

#include <sys/time.h>

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>

#include <math.h>
#include <string.h>

#include <pthread.h>
#include <sched.h>
#include <sys/mman.h>

#include "fweelin_fluidsynth.h"

#define fluid_sfont_iteration_start(_sf) (*(_sf)->iteration_start)(_sf)
#define fluid_sfont_iteration_next(_sf,_pr) (*(_sf)->iteration_next)(_sf,_pr)
#define fluid_preset_get_name(_preset) (*(_preset)->get_name)(_preset)
#define fluid_preset_get_banknum(_preset) (*(_preset)->get_banknum)(_preset)
#define fluid_preset_get_num(_preset) (*(_preset)->get_num)(_preset)

void FluidSynthParam_Int::Send(fluid_settings_t *settings) {
  printf("FLUID: Setting parameter '%s' = '%d'\n",name,val);
  fluid_settings_setint(settings, name, val);
};

void FluidSynthParam_Num::Send(fluid_settings_t *settings) {
  printf("FLUID: Setting parameter '%s' = '%f'\n",name,val);
  fluid_settings_setnum(settings, name, val);
};

void FluidSynthParam_Str::Send(fluid_settings_t *settings) {
  printf("FLUID: Setting parameter '%s' = '%s'\n",name,val);
  fluid_settings_setstr(settings, name, val);
};

FluidSynthProcessor::FluidSynthProcessor(Fweelin *app) : Processor(app),
  enable(1) {
  nframes_t bufsz = app->getBUFSZ();
  leftbuf = new sample_t[bufsz];
  rightbuf = new sample_t[bufsz];
  outbuf = new sample_t[bufsz];

  // Setup FluidSynth
  settings = new_fluid_settings();
  {
    FluidSynthParam *cur = app->getCFG()->GetFluidParam();
    while (cur != 0) {
      cur->Send(settings);
      cur = cur->next;
    }
  }
  // Set sampling rate
  fluid_settings_setnum(settings, "synth.sample-rate", 
			app->getAUDIO()->get_srate());  
  // Create synth
  synth = new_fluid_synth(settings);

  // Set interpolation quality
  printf("FLUID: Setting interpolation quality: %d\n",
	 app->getCFG()->GetFluidInterpolation());
  fluid_synth_set_interp_method(synth,-1,
				app->getCFG()->GetFluidInterpolation());

  // Load soundfonts
  {
    FluidSynthSoundFont *cur = app->getCFG()->GetFluidFont();
    while (cur != 0) {
      printf("FLUID: Loading SoundFont '%s'\n",cur->name);
      fluid_synth_sfload(synth,cur->name,1);
      cur = cur->next;
    }
  }

  // Setup patch names
  SetupPatches();

  app->getEMG()->ListenEvent(this,0,T_EV_FluidSynthEnable);
  app->getEMG()->ListenEvent(this,0,T_EV_FluidSynthAdjustPatch);

  // Listen for MIDI echo, and send it here
  FloConfig *fs = app->getCFG();
  app->getEMG()->ListenEvent(this,fs->GetInputMatrix(),
			     T_EV_Input_MIDIKey);
  app->getEMG()->ListenEvent(this,fs->GetInputMatrix(),
			     T_EV_Input_MIDIController);
  app->getEMG()->ListenEvent(this,fs->GetInputMatrix(),
			     T_EV_Input_MIDIPitchBend);
};

FluidSynthProcessor::~FluidSynthProcessor() {
  app->getEMG()->UnlistenEvent(this,0,T_EV_FluidSynthEnable);
  app->getEMG()->UnlistenEvent(this,0,T_EV_FluidSynthAdjustPatch);

  FloConfig *fs = app->getCFG();
  app->getEMG()->UnlistenEvent(this,fs->GetInputMatrix(),
			       T_EV_Input_MIDIKey);
  app->getEMG()->UnlistenEvent(this,fs->GetInputMatrix(),
			       T_EV_Input_MIDIController);
  app->getEMG()->UnlistenEvent(this,fs->GetInputMatrix(),
			       T_EV_Input_MIDIPitchBend);

  if (patches.patches != 0) {
    delete[] patches.patches;
    patches.patches = 0;
  }

  delete_fluid_synth(synth); 
  delete_fluid_settings(settings);

  delete[] leftbuf;
  delete[] rightbuf;
  delete[] outbuf;
};

void FluidSynthProcessor::process(char pre, nframes_t len, 
				  AudioBuffers *ab) {
  if (!pre) {
    // Map our output to the end of the external inputs
    ab->ins[ab->numins_ext] = outbuf;
    
    if (enable) {
      // Run synth DSP
      fluid_synth_write_float(synth, len, 
			      leftbuf, 0, 1,
			      rightbuf, 0, 1);
    
      // Now, combine the left and right channel into the output
      for (nframes_t l = 0; l < len; l++)
	outbuf[l] = (leftbuf[l] + rightbuf[l]) / 2;
    } else {
      // Silence, not enabled
      memset(outbuf,0,sizeof(sample_t) * len);
    }
  }
};

// Read new current patch from 'patches' and send patch change to synth
void FluidSynthProcessor::SendPatchChange() {
  if (patches.patches != 0) {
    if (patches.patchnum >= 0 && patches.patchnum < patches.numpatches) {
      FluidPatch *curp = &(patches.patches[patches.patchnum]);
      fluid_synth_program_select(synth, app->getCFG()->GetFluidChannel(), 
				 curp->sfontid, curp->bank, curp->prog);
    }
    else {
      printf("FLUID: Invalid current patch number!\n");
    }
  } else {
    printf("FLUID: Can't change patch- no soundfonts loaded!\n");
  }
};

// Sets up our internal patch list based on loaded soundfonts
void FluidSynthProcessor::SetupPatches() {
  fluid_preset_t preset;

  if (patches.patches != 0) {
    delete[] patches.patches;
    patches.patches = 0;
  }

  // Count patches
  int sfcnt = fluid_synth_sfcount(synth);
  patches.numpatches = 0;
  patches.patchnum = 0;
  for (int i = 0; i < sfcnt; i++) {
    fluid_sfont_t *curfont = fluid_synth_get_sfont(synth,i);
    fluid_sfont_iteration_start(curfont);
    while (fluid_sfont_iteration_next(curfont, &preset))
      patches.numpatches++;
  }

  // Allocate space for patch data
  if (patches.numpatches > 0) {
    patches.patches = new FluidPatch[patches.numpatches];
    
    // Store patch data
    int curp = 0;
    for (int i = 0; i < sfcnt; i++) {
      fluid_sfont_t *curfont = fluid_synth_get_sfont(synth,i);
      fluid_sfont_iteration_start(curfont);
      while (fluid_sfont_iteration_next(curfont, &preset)) {
	patches.patches[curp].sfontid = fluid_sfont_get_id(curfont);
	patches.patches[curp].bank = fluid_preset_get_banknum(&preset);
	patches.patches[curp].prog = fluid_preset_get_num(&preset);
	patches.patches[curp].name = fluid_preset_get_name(&preset);
	curp++;
      }
    }
  }
};

void FluidSynthProcessor::AdjustPatch (int patchofs, int fontofs) {
  if (patches.patches != 0) {
    int newp = patches.patchnum;
    if (fontofs != 0) {
      // Move by a whole soundfont
      int fontcnt = abs(fontofs);
      for (int i = 0; i < fontcnt; i++) {
	int curfont = patches.patches[newp].sfontid;
	while (newp >= 0 && newp < patches.numpatches && 
	       patches.patches[newp].sfontid == curfont)
	  newp = (fontofs > 0 ? newp+1 : newp-1);
      }
    }
    else {
      // Move patch by patch
      newp += patchofs;
    }

    // Check that new patch is within range
    if (newp < 0)
      newp = 0;
    else if (newp >= patches.numpatches)
      newp = patches.numpatches-1;
    
    patches.patchnum = newp;
    SendPatchChange();
  } else {
    printf("FLUID: Can't change patch because no patch to change!\n");
  }
}

void FluidSynthProcessor::ReceiveEvent(Event *ev, EventProducer *from) {
  FloConfig *fs = app->getCFG();

  switch (ev->GetType()) {
  case T_EV_FluidSynthEnable :
    {
      FluidSynthEnableEvent *fev = (FluidSynthEnableEvent *) ev;
      
      // OK!
      if (CRITTERS)
	printf("FLUID: Received FluidSynthEnable (%s)\n",
	       (fev->enable ? "on" : "off"));
      SetEnable(fev->enable);
    }
    break;

  case T_EV_FluidSynthAdjustPatch :
    {
      FluidSynthAdjustPatchEvent *fev = (FluidSynthAdjustPatchEvent *) ev;
      
      // OK!
      if (CRITTERS)
	printf("FLUID: Received FluidSynthAdjustPatch "
	       "(Patch: %d, SoundFont: %d)\n",
	       fev->patchadjust, fev->fontadjust);
      AdjustPatch(fev->patchadjust, fev->fontadjust);
    }
    break;

  case T_EV_Input_MIDIController :
    {
      MIDIControllerInputEvent *mcev = (MIDIControllerInputEvent *) ev;
      fluid_synth_cc(synth,app->getCFG()->GetFluidChannel(),
		     mcev->ctrl,mcev->val);
    }
    break;

  case T_EV_Input_MIDIPitchBend :
    {
      MIDIPitchBendInputEvent *mpev = (MIDIPitchBendInputEvent *) ev;
      fluid_synth_pitch_bend(synth,app->getCFG()->GetFluidChannel(),
			     mpev->val + FLUIDSYNTH_PITCHBEND_CENTER);
    }
    break;
      
  case T_EV_Input_MIDIKey :
    {
      MIDIKeyInputEvent *mkev = (MIDIKeyInputEvent *) ev;
      if (mkev->down)
	fluid_synth_noteon(synth,
			   app->getCFG()->GetFluidChannel(),
			   mkev->notenum+fs->transpose,
			   mkev->vel);
      else
	fluid_synth_noteoff(synth,
			    app->getCFG()->GetFluidChannel(),
			    mkev->notenum+fs->transpose);
    }
    break;

  default:
    break;
  }
}

#endif 
